name = "actinet"
__author__ = "Shing Chan, Aidan Acquah, Charilaos Zisou, Aiden Doherty"
__maintainer__ = "Shing Chan"
__maintainer_email__ = "shing.chan@ndph.ox.ac.uk"
__license__ = "See LICENSE file."

__model_version__ = "ssl_ukb_c24_rw_20240204"
__model_md5__ = "84f3d5bb73de5c4da057918c45400da4"

from . import _version

__version__ = _version.get_versions()["version"]
